# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define MetaProfiler

Use it to profile noWorkflow itself.
It is disabled by deault. To enabled, change active to True
"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import csv
import os
from collections import defaultdict
from datetime import datetime
from functools import wraps


class MetaProfiler(object):

    def __init__(self, active=False):
        self.file = 'nowtime.csv'
        self.active = active
        self.order = [
            "cmd",
            "definition",
            "deployment", "environment", "modules",
            "execution",
            "storage"
        ]
        self.data = defaultdict(float)

    def __call__(self, typ):
        def dec(f):
            if typ not in self.order:
                self.order.append(typ)
            @wraps(f)
            def wrapper(*args, **kwargs):
                before = datetime.now()
                result = f(*args, **kwargs)
                after = datetime.now()
                self.data[typ] += (after - before).total_seconds()
                return result
            return wrapper
        return dec

    def save(self):
        if self.active:
            row = [self.data[name] for name in self.order]
            rows = []
            if not os.path.exists(self.file):
                rows.append(self.order)
            rows.append(row)

            with open(self.file, 'a') as f:
                a = csv.writer(f)
                a.writerows(rows)

meta_profiler = MetaProfiler(active=False)



# Copyright (c) 2014 Universidade Federal Fluminense (UFF), Polytechnic Institute of New York University.
# This file is part of noWorkflow. Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import functools
from flask import render_template, jsonify, request
from ..persistence import persistence
from ..models.history import History
from ..models.trial import Trial
from ..models.diff import Diff

class WebServer(object):

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(WebServer, cls).__new__(
                                cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        from flask import Flask

        self.app = Flask(__name__)

app = WebServer().app

def connection(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        persistence.connect_existing(app.dir)
        return func(*args, **kwargs)
    return inner

@app.route('/<path:path>')
def static_proxy(path):
    return app.send_static_file(path)

@app.route('/trials')
@connection
def trials():
    history = History()
    return jsonify(**history.graph_data(request.args.get('script'),
                                        request.args.get('execution')))

@app.route('/')
@connection
def index():
    history = History()
    return render_template("index.html",
        cwd = os.getcwd(),
        scripts = history.scripts()
    )

@app.route('/<tid>-<graph_mode>')
@connection
def index2(tid, graph_mode):
    history = History()
    return render_template("index.html",
        cwd = os.getcwd(),
        scripts = history.scripts()
    )

@app.route('/trials/<tid>/<graph_mode>/<cache>.json')
@connection
def trial_graph(tid, graph_mode, cache):
    trial = Trial(tid)
    graph = trial.graph
    graph.use_cache &= bool(int(cache))
    finished, g = getattr(graph, graph_mode)(trial)
    return jsonify(**g)

@app.route('/trials/<tid>/dependencies')
@connection
def dependencies(tid):
    trial = Trial(tid)
    local, result = trial.modules()
    return jsonify(local=local, all=result)

@app.route('/trials/<tid>/all_modules')
@connection
def all_modules(tid):
    trial = Trial(tid)
    local, result = trial.modules()
    result = sorted(result, key=lambda x:x not in local)
    return render_template("trial.html",
        cwd = os.getcwd(),
        trial = trial.info(),
        modules = result,
        info = "modules.html",
    )

@app.route('/trials/<tid>/environment')
@connection
def environment(tid):
    trial = Trial(tid)
    return jsonify(env=trial.environment())


@app.route('/trials/<tid>/all_environment')
@connection
def all_environment(tid):
    trial = Trial(tid)
    return render_template("trial.html",
        cwd = os.getcwd(),
        trial = trial.info(),
        env = trial.environment(),
        info = "environment.html",
    )

@app.route('/trials/<tid>/file_accesses')
@connection
def file_accesses(tid):
    trial = Trial(tid)
    return jsonify(file_accesses=trial.file_accesses())

@app.route('/trials/<tid>/all_file_accesses')
@connection
def all_file_accesses(tid):
    trial = Trial(tid)
    return render_template("trial.html",
        cwd = os.getcwd(),
        trial = trial.info(),
        file_accesses = trial.file_accesses(),
        info = "file_accesses.html",
    )


@app.route('/diff/<trial1>/<trial2>/<tl>-<nh>-<graph_mode>')
@app.route('/diff/<trial1>/<trial2>')
@connection
def diff(trial1, trial2, tl=None, nh=None, graph_mode=None):
    diff = Diff(trial1, trial2)
    modules_added, modules_removed, modules_replaced = diff.modules()
    env_added, env_removed, env_replaced = diff.environment()
    fa_added, fa_removed, fa_replaced = diff.file_accesses()
    return render_template("diff.html",
        cwd = os.getcwd(),
        trial1 = diff.trial1.info(),
        trial2 = diff.trial2.info(),
        trial = diff.trial(),
        modules_added = modules_added,
        modules_removed = modules_removed,
        modules_replaced = modules_replaced,
        env_added = env_added,
        env_removed = env_removed,
        env_replaced = env_replaced,
        fa_added = fa_added,
        fa_removed = fa_removed,
        fa_replaced = fa_replaced,
    )

@app.route('/diff/<trial1>/<trial2>/<graph_mode>/<tl>-<nh>-<cache>.json')
@connection
def diff_graph(trial1, trial2, graph_mode, tl, nh, cache):
    diff = Diff(trial1, trial2)
    graph = diff.graph
    graph.use_cache &= bool(int(cache))
    finished, d, t1, t2 = getattr(graph, graph_mode)(
        diff, time_limit=int(tl), neighborhoods=int(nh))
    return jsonify(
        diff=d,
        trial1=t1,
        trial2=t2,
    )



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define utility functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


from .consts import FORMAT
from .data import concat_iter, OrderedCounter, hashabledict
from .functions import wrap, resource, calculate_duration
from .io import redirect_output, print_msg, print_fn_msg, verbose
from .metaprofiler import meta_profiler



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" 'now restore' command """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
from datetime import datetime

from .command import Command
from .. import prov_deployment
from ..cross_version import items
from ..models.trial import Trial
from ..persistence import persistence
from ..utils.io import print_msg


class Restore(Command):
    """ Restore the files of a trial """

    def __init__(self, *args, **kwargs):
        super(Restore, self).__init__(*args, **kwargs)
        self.print_msg = True

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('trial', type=int, nargs='?',
                help='trial id or none for last trial')
        add_arg('-s', '--script',
                help='python script to be restored')
        add_arg('-b', '--bypass-modules', action='store_true',
                help='bypass module dependencies analysis, assuming that no '
                     'module changes occurred since last execution')
        add_arg('-l', '--local', action='store_true',
                help='restore local modules')
        add_arg('-i', '--input', action='store_true',
                help='restore input files')
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def create_backup(self, trial, args):
        """ Create a backup trial if the script were changed """
        if not os.path.isfile(trial.script):
            return

        head = trial.head_trial()
        with open(trial.script, 'rb') as fil:
            code = fil.read()
            code_hash = persistence.put(code)

        if code_hash != head.code_hash:
            now = datetime.now()
            tid = persistence.store_trial(
                now, trial.script, code,
                '<restore {}>'.format(trial.id),
                args.bypass_modules, run=False)
            prov_deployment.collect_provenance(args, {
                'trial_id': tid,
                'code': code,
                'path': trial.script,
                'compiled': None,
            })
            print_msg('Backup Trial {} created'.format(tid), self.print_msg)

    def restore(self, path, code_hash, trial_id):
        """ Restore file with <code_hash> from <trial_id> """
        load_file = persistence.get(code_hash)
        with open(path, 'wb') as fil:
            fil.write(load_file)
        print_msg('File {} from trial {} restored'.format(path, trial_id),
                  self.print_msg)

    def restore_script(self, trial):
        """ Restore the main script from <trial> """
        self.restore(trial.script, trial.code_hash, trial.id)

    def execute(self, args):
        persistence.connect_existing(args.dir or os.getcwd())
        trial = Trial(trial_id=args.trial, script=args.script, exit=True)
        self.create_backup(trial, args)

        self.restore_script(trial)

        persistence.store_parent(trial.script, trial.id)
        if args.local:
            local_modules, _ = trial.modules()
            for module in local_modules:
                self.restore(module['path'], module['code_hash'], trial.id)

        if args.input:
            file_accesses = trial.file_accesses()
            files = {}
            for faccess in reversed(file_accesses):
                files[faccess['name']] = faccess['content_hash_before']
            for name, content_hash in items(files):
                self.restore(name, content_hash, trial.id)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Test now.prov_deployment module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import unittest
import modulefinder
import platform
import os
from ..now.persistence import persistence
from ..now.prov_deployment import collect_environment_provenance
from ..now.prov_deployment import collect_modules_provenance
from ..now.prov_deployment import get_version
from . import mock

persistence = mock.Mock()


class TestProvDeployment(unittest.TestCase):
    """TestCase for now.prov_deployment module"""

    def test_collect_environment_provenance(self):
        env = collect_environment_provenance()
        self.assertIn('PWD', env)
        self.assertIn('PYTHON_VERSION', env)
        self.assertIn('PYTHON_IMPLEMENTATION', env)

    def _test_collect_modules_provenance(self):
        finder = modulefinder.ModuleFinder()
        finder.run_script(__file__)
        modules = collect_modules_provenance(finder.modules)
        modules = {m[0] for m in modules}
        self.assertIn('modulefinder', modules)

    def test_get_version_system_module(self):
        self.assertEqual(platform.python_version(), get_version('sys'))

    def test_get_version_distribution(self):
        self.assertEqual('1.0.2', get_version('pyposast'))

    def test_get_version_other(self):
        self.assertEqual(None, get_version('other'))



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Command base """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from ..utils.functions import abstract


class Command(object):
    """ Command base """
    def __init__(self, cmd=None):
        self.cmd = cmd or type(self).__name__.lower()
        self.help = self.__doc__.split('\n')[0].strip()
        self.parser = None

    def create_parser(self, subparsers):
        """ Create parser with arguments """
        kwargs = {}
        if self.help:
            kwargs['help'] = self.help
        self.parser = subparsers.add_parser(self.cmd, **kwargs)

        self.add_arguments()
        self.parser.set_defaults(func=self.execute)

    def add_arguments(self):
        """ Add arguments to command. Override on subclass """
        pass

    def add_argument(self, *args, **kwargs):
        """ Add argument to parser available for both IPython magic and cmd """
        return self.parser.add_argument(*args, **kwargs)

    def add_argument_cmd(self, *args, **kwargs):
        """ Add argument to parser available only for cmd """
        return self.parser.add_argument(*args, **kwargs)

    def execute(self, args):
        """ Execute the command. Override on subclass """
        abstract()
        print(self, args)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

class Mock(object):

    def __getattr__(self, name):
        print(name)
        return lambda *args, **kwargs: None



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Define Metadata classes to be used as configuration """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import fnmatch
import os
import sys

from datetime import datetime
from .persistence import persistence, get_serialize
from .prov_definition.definition import Definition
from .utils.io import print_msg


LAST_TRIAL = ".last_trial"
MAIN = 0
PACKAGE = 1
ALL = 2

CONTEXTS = {
    'main': MAIN,
    'package': PACKAGE,
    'all': ALL
}


class RunMetascript(object):

    def __init__(self):
        # Trial id read from Database : int
        self.trial_id = None
        # Trial name : str
        self.name = None
        # Source code of the main script : str
        self.code = None
        # Compiled code : types.CodeType
        self.compiled = None
        # Script dir : str
        self.dir = None
        # Main path : str
        self._path = None
        # List of paths considered : set(str)
        self.paths = set(str())
        # Main namespace : dict
        self.namespace = None
        # Argv
        self.argv = None
        # Object Serialize function
        self.serialize = None

        # Verbose print
        self.verbose = False
        # Should it create a file with the last executed trial id : bool
        self.should_create_last_file = False
        # Profile noWorkflow itself
        self.meta = False

        # Definition object : Definition
        self.definition = Definition()
        # Show script disassembly : bool
        self.disasm = False
        # Show script disassembly before changes : bool
        self.disasm0 = False

        # Bypass module check : bool
        self.bypass_modules = False

        # Depth for capturing function activations : int
        self.depth = sys.getrecursionlimit()
        # Depth for capturing function activations outside context : int
        self.non_user_depth = 1
        # Execution provenance provider : str
        self.execution_provenance = "Profiler"
        # Script context : ['main', 'package', 'all']
        self._context = MAIN
        # Save every X ms : int
        self.save_frequency = 1000
        # Save after closing X activations
        self.call_storage_frequency = 0

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    @property
    def path(self):
        """ Return path """
        return self._path

    @path.setter
    def path(self, path):
        """ Set _path.
            Remove old path from paths list and add new one to it """
        if self._path:
            self.paths.remove(self._path)
        self._path = path
        if not path:
            return
        with open(path, "rb") as script_file:
            self.code = script_file.read()
        self.paths.add(path)
        self.dir = self.dir or os.path.dirname(path)

    def add_path(self, path):
        """ Add path to paths list """
        self.paths.add(path)

    @property
    def context(self):
        """ Return context """
        return self._context

    @context.setter
    def context(self, context):
        """ Set context
            Must be set AFTER self.non_user_depth """
        self._context = CONTEXTS[context]
        if self._context in (PACKAGE, ALL):
            dirname = os.path.dirname(self.path)
            for root, _, filenames in os.walk(dirname):
                for filename in fnmatch.filter(filenames, "*.py"):
                    self.add_path(os.path.join(root, filename))
        if context == ALL:
            self.non_user_depth = self.depth

    def clear_namespace(self, erase=True):
        """ Clear namespace dict """
        if erase:
            self.namespace.clear()
        self.namespace.update({"__name__"    : "__main__",
                               "__file__"    : self.path,
                               "__builtins__": __builtins__,
                              })

    def clear_sys(self):
        """ Clear sys variables """
        # Replace now's dir with script's dir in front of module search path.
        sys.path[0] = self.dir
        # Clear argv
        sys.argv = self.argv

    def read_cmd_args(self, args):
        """ Read cmd line argument object """
        self.dir = args.dir
        self.argv = args.argv
        self.path = args.script
        self.should_create_last_file = args.create_last
        self.name = args.name or os.path.basename(args.argv[0])
        return self._read_args(args)

    def read_ipython_args(self, args, directory, filename, argv, create_last):
        """ Read magic line argument object """
        self.dir = directory
        self.argv = argv
        self.path = filename
        self.should_create_last_file = create_last
        self.name = args.name or os.path.basename(filename)
        return self._read_args(args)

    def _read_args(self, args):
        """ Read cmd line argument object """
        self.serialize = get_serialize(args) # ToDo: add serializer param
        self.verbose = args.verbose
        self.meta = args.meta

        self.disasm = args.disasm
        self.disasm0 = args.disasm0
        self.bypass_modules = args.bypass_modules

        self.depth = args.depth
        self.non_user_depth = args.non_user_depth
        self.execution_provenance = args.execution_provenance
        self.context = args.context
        self.save_frequency = args.save_frequency
        self.call_storage_frequency = args.call_storage_frequency
        return self

    def create_last(self):
        """ Create file indicating last trial id """
        if self.should_create_last_file:
            lastname = os.path.join(os.path.dirname(self.path), LAST_TRIAL)
            with open(lastname, "w") as lastfile:
                lastfile.write(str(self.trial_id))

    def create_trial(self):
        """ Create trial and assign a new id to it """
        now = datetime.now()
        try:
            self.trial_id = persistence.store_trial(
                now, self.name, self.code, " ".join(sys.argv[1:]),
                self.bypass_modules)
        except TypeError as e:
            if self.bypass_modules:
                print_msg("not able to bypass modules check because no "
                          "previous trial was found", True)
                print_msg("aborting execution", True)
            else:
                raise e

            sys.exit(1)



#!/usr/bin/env python
from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop

import fnmatch
import os
import platform

from uuid import getnode

__version__ = "0.12.3"


def recursive_path(pack, path):
    matches = []
    for root, dirnames, filenames in os.walk(os.path.join(pack, path)):
        for filename in filenames:
            matches.append(os.path.join(root, filename)[len(pack) + 1:])
    return matches


try:
   import pypandoc
   long_description = pypandoc.convert('../README.md', 'rst')
except (IOError, ImportError):
   long_description = (
        "Supporting infrastructure to run scientific experiments "
        "without a scientific workflow management system.")


def analytics(command_subclass):
    """Sends install statistics to analytics."""
    orig_run = command_subclass.run

    def modified_run(self):
        import sys
        try:
            try:
                from urllib2 import HTTPHandler, build_opener
                from urllib2 import urlopen, Request
                from urllib import urlencode
            except ImportError:
                from urllib.request import HTTPHandler, build_opener
                from urllib.request import urlopen, Request
                from urllib.parse import urlencode

            os_ver = platform.system()
            py_ver = '_'.join(str(x) for x in sys.version_info)
            now_ver = __version__.replace('.', '_')

            code = 'os:{0},py:{1},now:{2}'.format(os_ver, py_ver, now_ver)
            action = command_subclass.action
            cid = getnode()
            payload = {
                'v': '1',
                'tid': 'UA-61791314-1',
                'cid': str(cid),
                't': 'event',
                'ec': action,
                'ea': code,
            }

            url = 'http://www.google-analytics.com/collect'
            data = urlencode(payload).encode('utf-8')
            request = Request(url, data=data)
            request.get_method = lambda: "POST"
            connection = urlopen(request)
        except:
            pass
        orig_run(self)

    command_subclass.run = modified_run
    return command_subclass


@analytics
class CustomDevelopCommand(develop):
    action = 'develop'


@analytics
class CustomInstallCommand(install):
    action = 'install'


setup(
    name = "noworkflow",
    version = __version__,
    packages = find_packages(),
    package_data = {
        'noworkflow': [
            'resources/*',
        ] + recursive_path('noworkflow', 'now/vis/static')
          + recursive_path('noworkflow', 'now/vis/templates'),
    },
    entry_points = {'console_scripts': ['now = noworkflow.main:main']},
    author = ("Joao Pimentel, Leonardo Murta, Vanessa Braganholo, "
              "Fernando Chirigati, David Koop, and Juliana Freire"),
    author_email = "leomurta@ic.uff.br",
    description = "Supporting infrastructure to run scientific experiments "
                  "without a scientific workflow management system.",
    long_description = long_description,
    license = "MIT",
    keywords = "scientific experiments provenance python",
    url = "https://github.com/gems-uff/noworkflow",
    install_requires=['pyposast', 'jsonpickle'],
    extras_require = {
        'vis': ['pyposast', 'flask'],
        'notebook': ['pyposast', 'ipython', 'jupyter'],
        'all': ['pyposast', 'ipython', 'jupyter', 'flask', 'pyswip-alt'],
    },
    cmdclass={
        'install': CustomInstallCommand,
        'develop': CustomDevelopCommand,
    },
)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys

from IPython.utils.process import arg_split
from IPython.core import magic_arguments
from IPython.core.magic import  line_magic, cell_magic, line_cell_magic

from ...cmd import Command


MAGIC_TYPES = {
    'cell': cell_magic,
    'line': line_magic,
    'line_cell': line_cell_magic
}


class IpythonCommandMagic(Command):

    def __init__(self, magic, docstring, magic_type='cell'):
        self.__doc__ = docstring
        super(IpythonCommandMagic, self).__init__()
        self.magic = magic
        self.docstring = docstring
        self.magic_type = magic_type
        self.args = []

    def add_argument_cmd(self, *args, **kwargs):
        pass

    def add_argument(self, *args, **kwargs):
        self.args.append(
            magic_arguments.argument(*args, **kwargs)
        )

    def create_magic(self, f):
        f.__name__ = str(self.magic)
        f.__doc__ = self.docstring
        f = MAGIC_TYPES[self.magic_type](self.magic)(f)
        for arg in self.args:
            f = arg(f)
        f = magic_arguments.magic_arguments()(f)
        return f

    def execute(self, func, line, cell, magic_cls):
        super(Command, self).execute(args)

    def arguments(self, func, line):
        argv = arg_split(line, posix = not sys.platform.startswith('win'))
        args = magic_arguments.parse_argstring(func, line)
        return argv, args



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Program Slicing Tracer Module """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
import dis
import linecache
import traceback
from datetime import datetime
from functools import wraps, partial
from opcode import HAVE_ARGUMENT, EXTENDED_ARG
from inspect import ismethod

from .data_objects import ObjectStore, Variable, Dependency, Usage, Return
from .profiler import Profiler
from ..utils import io
from ..utils.io import print_fn_msg
from ..utils.bytecode.f_trace import find_f_trace, get_f_trace
from ..cross_version import items, IMMUTABLE
from ..prov_definition import ClassDef, Assert, With, Decorator

from ..persistence import persistence


WITHOUT_PARAMS = (ClassDef, Assert, With)


class JointPartial(partial):

    def __init__(self):
        super(JointPartial, self).__init__()
        self._joint_tracer = True

    def __repr__(self):
        return "Joint<{}>".format(', '.join("{}.{}".format(
            (a.im_self if ismethod(a) else a).__class__.__name__, a.__name__)
            for a in self.args))

    @property
    def __name__(self):
        return "Joint"


def joint_tracer(first, second, frame, event, arg):
    """  Joint tracer used to combine noWorkflow tracer with other tracers """
    try:
        first = first(frame, event, arg)
    except:
        traceback.print_exc()
    try:
        second = second(frame, event, arg)
    except:
        traceback.print_exc()
    return create_joint_tracer(first, second)


def create_joint_tracer(first, second):
    """  Create joint tracer """
    if not first:
        return second
    if not second:
        return first
    if first == second:
        return first
    joint = partial(joint_tracer, first, second)
    joint._joint_tracer = True
    #joint.__str__ = lambda x: "Joint<{}, {}>".format(first, second)

    return joint

_sys_settrace = sys.settrace

def joint_settrace(tracer):
    current = sys.gettrace()
    if hasattr(current, '_joint_tracer'):
        new = create_joint_tracer(current.args[0], tracer)
    elif current:
        new = create_joint_tracer(current, tracer)
    else:
        new = tracer
    _sys_settrace(new)

sys.settrace = joint_settrace


class Tracer(Profiler):
    """ Tracer used for program slicing """
    # pylint: disable=R0902

    def __init__(self, *args):
        super(Tracer, self).__init__(*args)
        definition = self.metascript.definition
        self.event_map['line'] = self.trace_line
        self.event_map['exception'] = self.trace_exception

        # Store slicing provenance
        self.variables = ObjectStore(Variable)
        self.dependencies = ObjectStore(Dependency)
        self.usages = ObjectStore(Usage)

        # Returns
        self.returns = {}
        # Last iter
        self.last_iter = None
        # Next is iteration
        self.next_is_iter = False

        # Useful maps
        # Map of dependencies by line
        self.line_dependencies = definition.line_dependencies
        # Map of dependencies by line
        self.line_gen_dependencies = definition.line_gen_dependencies
        # Map of name_refs by line
        self.line_usages = definition.line_usages
        # Map of calls by line and col
        self.call_by_col = definition.call_by_col
        # Map of calls by line and lasti
        self.call_by_lasti = definition.call_by_lasti
        # Map of with __enter__ by line and lasti
        self.with_enter_by_lasti = definition.with_enter_by_lasti
        # Map of with __exit__ by line and lasti
        self.with_exit_by_lasti = definition.with_exit_by_lasti
        # Set of imports
        self.imports = definition.imports
        # Set of GET_ITER and FOR_ITER lasti by line
        self.iters = definition.iters

        # Allow debuggers:
        self.f_trace_frames = []

        # Events are not unique. Profiler and Tracer have same events
        self.unique_events = False

    def remove_return_lasti(self, lasti_set):
        """ Remove lasti from list of returns """
        returns = self.returns
        for lasti in lasti_set:
            del returns[lasti]

    def add_variable(self, name, line, f_locals, value='--check--'):
        """ Adds variable """
        if value == '--check--' and name in f_locals:
            value = self.serialize(f_locals[name])
        else:
            value = 'now(n/a)'
        return self.variables.add(name, line, value, datetime.now())

    def add_dependency(self, var, dep, activation, filename, lasti_set):
        """ Adds dependency """
        # pylint: disable=R0913
        var_id, dependencies_add = var.id, self.dependencies.add
        context = activation.context

        # Variable
        if dep in context:
            dependencies_add(var_id, context[dep].id)
        # Function Call
        if isinstance(dep, tuple):
            call = self.call_by_col[filename][dep[0]][dep[1]]
            lasti, returns = call.lasti, self.returns

            if not lasti in returns:
                return
            _return = returns[lasti]

            if isinstance(_return, Return):
                # Call was not evaluated before
                vid = self.add_variable('call {}'.format(
                    _return.activation.name), dep[0], {}, 'now(n/a)')
                returns[lasti] = (_return, vid)
                call.result = (_return.var.id, _return.var.line)
                dependencies_add(vid, _return.var.id)

                # Remove lasti from returns list to avoid conflict
                lasti_set.add(lasti)
            else:
                _return, vid = _return

            self.add_dependencies(self.variables[vid], activation, call.func,
                                  filename, lasti_set)
            dependencies_add(var_id, vid)
        # Iter
        if dep == 'now(iter)' and self.last_iter:
            dependencies_add(var_id, self.last_iter)

    def add_dependencies(self, var, activation, dependencies, filename,
                         lasti_set):
        """ Adds dependencies to var """
        # pylint: disable=R0913
        add_dependency = self.add_dependency
        for dep in dependencies:
            add_dependency(var, dep, activation, filename, lasti_set)

    def slice_line(self, activation, lineno, f_locals, filename,
                   line_dependencies=None):
        """ Generates dependencies from line """
        # pylint: disable=R0913
        # pylint: disable=R0914
        if line_dependencies is None:
            line_dependencies = self.line_dependencies
        print_fn_msg(lambda: 'Slice [{}] -> {}'.format(
            lineno, linecache.getline(filename, lineno).strip()))
        context, variables = activation.context, self.variables
        usages_add = self.usages.add
        add_variable = self.add_variable
        add_dependencies = self.add_dependencies

        dependencies = line_dependencies[filename][lineno]
        for ctx in ('Load', 'Del'):
            usages = self.line_usages[filename][lineno][ctx]
            for name in usages:
                if name in context:
                    usages_add(context[name].id, name, lineno, ctx)

        lasti_set = set()
        set_last_iter = False
        for name, others in items(dependencies):
            if name == 'return':
                vid = add_variable(name, lineno, f_locals,
                                   value=activation.return_value)
                self.returns[activation.lasti] = Return(activation,
                                                        variables[vid])
            elif name == 'yield':
                vid = add_variable(name, lineno, f_locals)
                self.last_iter = vid
                set_last_iter = True
            else:
                vid = add_variable(name, lineno, f_locals)

            add_dependencies(
                variables[vid], activation, others, filename, lasti_set)

            context[name] = variables[vid]

        self.remove_return_lasti(lasti_set)
        if not set_last_iter:
            self.last_iter = None

    def close_activation(self, event, arg):
        """ Slice all lines from closing activation """
        for line in self.current_activation.slice_stack:
            self.slice_line(*line)
        super(Tracer, self).close_activation(event, arg)

    def add_generic_return(self, activation, frame, ccall=False):
        """ Add return to functions that do not have return
            For ccall, add dependency from all params
        """

        lineno = frame.f_lineno
        variables = self.variables

        # Artificial return condition
        lasti = activation.lasti
        if 'return' not in activation.context:
            vid = self.add_variable('return', lineno, {},
                                    value=activation.return_value)
            activation.context['return'] = variables[vid]
            self.returns[lasti] = Return(activation, variables[vid])

            filename = frame.f_code.co_filename
            if ccall and filename in self.paths:
                caller = self.current_activation
                call = self.call_by_lasti[filename][lineno][lasti]
                all_args = call.all_args()
                lasti_set = set()
                self.add_dependencies(variables[vid], caller, all_args,
                                      filename, lasti_set)
                self.add_inter_dependencies(frame, all_args, caller, lasti_set)
                self.remove_return_lasti(lasti_set)

    def match_arg(self, passed, arg, caller, activation, line, f_locals,
                  filename):
        """ Matches passed param with argument """
        # pylint: disable=R0913
        context = activation.context

        if arg in context:
            act_var = context[arg]
        else:
            vid = self.add_variable(arg, line, f_locals)
            act_var = self.variables[vid]
            context[arg] = act_var

        if passed:
            lasti_set = []
            self.add_dependency(act_var, passed, caller, filename, lasti_set)
            self.remove_return_lasti(lasti_set)

    def match_args(self, args, param, caller, activation, line, f_locals,
                   filename):
        """ Matches passed param with arguments """
        # pylint: disable=R0913
        for arg in args:
            self.match_arg(arg, param, caller, activation, line, f_locals,
                           filename)

    def add_inter_dependencies(self, frame, args, activation, lasti_set):
        """ Adds dependencies between params """
        # pylint: disable=R0914
        f_locals, context = frame.f_locals, activation.context
        lineno, variables = frame.f_lineno, self.variables
        add_variable = self.add_variable
        add_dependencies = self.add_dependencies

        added = {}
        for arg in args:
            try:
                var = f_locals[arg]
                if not isinstance(var, IMMUTABLE):
                    vid = add_variable(arg, lineno, {}, value=var)
                    variable = variables[vid]
                    add_dependencies(variable, activation, args,
                                     frame.f_code.co_filename, lasti_set)
                    added[arg] = variable
            except KeyError:
                pass

        for arg, variable in items(added):
            context[arg] = variable

    def add_argument_variables(self, frame):
        """ Adds argument variables """
        # pylint: disable=R0914
        back = frame.f_back
        fname = frame.f_code.co_name
        filename = back.f_code.co_filename
        lineno = back.f_lineno
        lasti = back.f_lasti
        if (fname == '__enter__' and
                lasti in self.with_enter_by_lasti[filename][lineno]):
            return
        if (fname == '__exit__' and
                lasti in self.with_exit_by_lasti[filename][lineno]):
            return
        if lasti in self.iters[filename][lineno]:
            self.next_is_iter = True
            return

        call = self.call_by_lasti[filename][lineno][lasti]
        if isinstance(call, WITHOUT_PARAMS):
            return
        if isinstance(call, Decorator) and not call.fn:
            return
        f_locals = frame.f_locals
        match_args = self.match_args
        match_arg = self.match_arg
        caller, act = self.parent_activation, self.current_activation
        act_args_index = act.args.index
        line = frame.f_lineno
        sub = -[bool(act.starargs), bool(act.kwargs)].count(True)

        order = act.args + act.starargs + act.kwargs
        used = [0 for _ in order]
        j = 0
        for i, call_arg in enumerate(call.args):
            j = i if i < len(order) + sub else sub
            act_arg = order[j]
            match_args(call_arg, act_arg, caller, act, line, f_locals,
                       filename)
            used[j] += 1
        for act_arg, call_arg in items(call.keywords):
            try:
                i = act_args_index(act_arg)
                match_args(call_arg, act_arg, caller, act, line, f_locals,
                           filename)
                used[i] += 1
            except ValueError:
                for kwargs in act.kwargs:
                    match_args(call_arg, kwargs, caller, act, line, f_locals,
                               filename)

        # ToDo: improve matching
        #   Ignore default params
        #   Do not match f(**kwargs) with def(*args)
        args = [(i, order[i]) for i in range(len(used)) if not used[i]]
        for star in call.kwargs + call.starargs:
            for i, act_arg in args:
                match_args(star, act_arg, caller, act, line, f_locals,
                           filename)
                used[i] += 1

        args = [(i, order[i]) for i in range(len(used)) if not used[i]]
        for i, act_arg in args:
            match_arg(None, act_arg, caller, act, line, f_locals, filename)

        lasti_set = set()
        self.add_inter_dependencies(back, call.all_args(), caller, lasti_set)
        self.remove_return_lasti(lasti_set)

    def trace_call(self, frame, event, arg):
        self.check_f_trace(frame, event)
        super(Tracer, self).trace_call(frame, event, arg)
        filename = frame.f_back.f_code.co_filename
        if (filename not in self.paths or not self.valid_depth() or
                frame.f_back.f_lineno in self.imports[filename]):
            return
        self.add_argument_variables(frame)

    def trace_c_return(self, frame, event, arg):
        self.check_f_trace(frame, event)
        activation = self.current_activation
        super(Tracer, self).trace_c_return(frame, event, arg)
        self.add_generic_return(activation, frame, ccall=True)
        activation.context['return'].value = activation.return_value

    def trace_return(self, frame, event, arg):
        self.check_f_trace(frame, event)
        activation = self.current_activation
        super(Tracer, self).trace_return(frame, event, arg)
        self.add_generic_return(activation, frame)
        activation.context['return'].value = activation.return_value

    def trace_line(self, frame, event, arg):
        """ Trace Line event """
        # pylint: disable=W0613
        self.check_f_trace(frame, event)
        code = frame.f_code
        filename = frame.f_code.co_filename
        lineno = frame.f_lineno

        loc, glob = frame.f_locals, frame.f_globals
        if find_f_trace(code, loc, glob, frame.f_lasti):
            _frame = get_f_trace(code, loc, glob)
            if _frame.f_trace:
                self.f_trace_frames.append(_frame)


        # Different file
        if filename not in self.paths:
            return

        activation = self.current_activation

        print_fn_msg(lambda: '[{}] -> {}'.format(
            lineno, linecache.getline(filename, lineno).strip()))
        if activation.slice_stack:
            self.slice_line(*activation.slice_stack.pop())
        if self.next_is_iter:
            self.slice_line(activation, lineno, frame.f_locals, filename,
                            line_dependencies=self.line_gen_dependencies)
            self.next_is_iter = False
        activation.slice_stack.append([
            activation, lineno, frame.f_locals, filename])

    def trace_c_exception(self, frame, event, arg):
        self.check_f_trace(frame, event)
        super(Tracer, self).trace_c_exception(frame, event, arg)

    def trace_exception(self, frame, event, arg):
        self.check_f_trace(frame, event)

    def check_f_trace(self, frame, event):
        if self.f_trace_frames:

            for _frame in self.f_trace_frames:
                _frame.f_trace = create_joint_tracer( _frame.f_trace, self.tracer)

            self.f_trace_frames = []
            return

    def tearup(self):
        _sys_settrace(self.tracer)
        sys.setprofile(self.tracer)

    def teardown(self):
        super(Tracer, self).teardown()
        _sys_settrace(self.default_trace)

    def store(self, partial=False):
        if not partial:
            while len(self.activation_stack) > 1:
                self.close_activation('store', None)
        super(Tracer, self).store(partial=partial)
        tid = self.trial_id
        persistence.store_slicing_variables(tid, self.variables, partial)
        persistence.store_slicing_dependencies(tid, self.dependencies, partial)
        persistence.store_slicing_usages(tid, self.usages, partial)


        #self.view_slicing_data(io.verbose)

    def view_slicing_data(self, show=True):
        """ View captured slicing """
        # pylint: disable=W0640
        # pylint: disable=W0631
        if show:
            for var in self.variables:
                print_fn_msg(lambda: var)

            for dep in self.dependencies:
                print_fn_msg(lambda: "{}\t<-\t{}".format(
                    self.variables[dep.dependent],
                    self.variables[dep.supplier]))

            for var in self.usages:
                print_fn_msg(lambda: var)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Collect deployment provenance from Python script"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import importlib
import modulefinder
import os
import platform
import socket
import sys
import pkg_resources

from .persistence import persistence
from .utils import print_msg, redirect_output, meta_profiler
from .cross_version import string, default_string, items


@meta_profiler("environment")
def collect_environment_provenance():
    """Collect enviroment variables and operating system characteristics
    Return dict
    """
    environment = {}
    e = environment

    environment['OS_NAME'] = platform.system()
    # Unix environment
    try:
        for name in os.sysconf_names:
            try:
                environment[name] = os.sysconf(name)
            except ValueError:
                pass
        for name in os.confstr_names:
            environment[name] = os.confstr(name)
        environment['USER'] = os.getlogin()
        e['OS_NAME'], _, e['OS_RELEASE'], e['OS_VERSION'], _ = os.uname()
    except:
        pass
    # Windows environment
    if environment['OS_NAME'] == 'Windows':
        e['OS_RELEASE'], e['OS_VERSION'], _, _ = platform.win32_ver()

    # Both Unix and Windows
    environment.update(os.environ)
    environment['PWD'] = os.getcwd()
    environment['PID'] = os.getpid()
    environment['HOSTNAME'] = socket.gethostname()
    environment['ARCH'] = platform.architecture()[0]
    environment['PROCESSOR'] = platform.processor()
    environment['PYTHON_IMPLEMENTATION'] = platform.python_implementation()
    environment['PYTHON_VERSION'] = platform.python_version()
    return environment

@meta_profiler("modules")
def collect_modules_provenance(modules):
    """Return a set of module dependencies in the form:
        (name, version, path, code_hash)
    Store module provenance in the persistence database
    """
    dependencies = []
    for name, module in items(modules):
        if name != '__main__':
            version = get_version(name)
            path = module.__file__
            if path == None:
                code_hash = None
            else:
                with open(path, 'rb') as f:
                    code_hash = persistence.put(f.read())
            dependencies.append((name, version, path, code_hash))
    return sorted(dependencies)


def get_version(module_name):
    """Get module version"""
    # Check built-in module
    if module_name in sys.builtin_module_names:
        return platform.python_version()

    # Check package declared module version
    try:
        # TODO: This is slow! Is there any alternative?
        return pkg_resources.get_distribution(module_name).version
    except:
        pass

    # Check explicitly declared module version
    try:
        module = importlib.import_module(module_name)
        for attr in ['__version__', 'version', '__VERSION__', 'VERSION']:
            try:
                version = getattr(module, attr)
                if isinstance(version, string):
                    return default_string(version)
                if isinstance(version, tuple):
                    return '.'.join(map(str, version))

            except AttributeError:
                pass
    except:
        pass

    # If no other option work, return None
    return None

@meta_profiler("deployment")
def collect_provenance(metascript):
    """Collect deployment provenance:
        - environment variables
        - modules dependencies

    metascript should have 'trial_id' and 'path'
    """
    print_msg('  registering environment attributes')
    environment = collect_environment_provenance()
    persistence.store_environment(metascript['trial_id'], environment)

    if metascript.bypass_modules:
        print_msg('  using previously detected module dependencies '
                  '(--bypass-modules).')
    else:
        print_msg('  searching for module dependencies')
        modules = []
        with redirect_output():
            excludes = set()
            last_name = None
            max_atemps = 1000
            for i in range(max_atemps):
                try:
                    finder = modulefinder.ModuleFinder(excludes=excludes)
                    finder.run_script(metascript['path'])
                    break
                except SyntaxError as e:
                    name = e.filename.split('site-packages/')[-1]
                    name = name.replace(os.sep, '.')
                    name = name[:name.rfind('.')]
                    if last_name is not None and last_name in name:
                        last_name = last_name[last_name.find('.') + 1:]
                    else:
                        last_name = name
                    excludes.add(last_name)
                    print_msg('    skip module due syntax error: {} ({}/{})'
                        .format(last_name, i+1, max_atemps))

            print_msg('  registering provenance from {} modules'.format(
                len(finder.modules) - 1))

            modules = collect_modules_provenance(finder.modules)
        persistence.store_dependencies(metascript['trial_id'], modules)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Intermediate Tree Structures and Graph Structures """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


import json
import time
import traceback
from collections import OrderedDict
from datetime import datetime
from sqlite3 import OperationalError

from ..cross_version import default_string, pickle, items
from ..persistence import persistence
from ..utils import calculate_duration, FORMAT, OrderedCounter, print_msg


class Graph(object):
    """ Graph superclass
        Handle json transformation"""
    # pylint: disable=R0201
    # pylint: disable=R0903
    def escape_json(self, data):
        """ Escape JSON """
        data = json.dumps(data)
        return (data.replace('&', '\\u0026')
                .replace('<', '\\u003c')
                .replace('>', '\\u003e'))


def prepare_cache(get_type):
    """ Decorator: Load graph from cache """
    def cache(name, attrs=""):
        """ Decorator: Load graph from cache """
        def dec(func):
            """ Decorator: Load graph from cache """
            def load(self, *args, **kwargs):
                """ Load graph from cache """
                typ = get_type(self, *args, **kwargs)
                attributes = ' '.join(
                    [str(kwargs[a]) for a in attrs.split() if a in kwargs])
                conditions = {
                    'type': '"{}"'.format(typ),
                    'name': '"{}"'.format(name),
                    'attributes': '"{}"'.format(attributes),
                }
                if self.use_cache:
                    try:
                        graph_cache = persistence.load('graph_cache',
                                                       **conditions)
                        for cache in graph_cache:
                            result = pickle.loads(persistence.get(
                                cache[default_string('content_hash')]))
                            if not result[0]:
                                continue
                            return result
                    except OperationalError:
                        traceback.print_exc()
                        print_msg("Couldn't load graph cache", True)
                start = time.time()
                graph = func(self, *args, **kwargs)
                duration = time.time() - start
                try:
                    persistence.delete('graph_cache', **conditions)
                    persistence.insert('graph_cache', {
                        'type': typ,
                        'name': name,
                        'duration': duration,
                        'attributes': attributes,
                        'content_hash': persistence.put(pickle.dumps(graph)),
                    })
                except OperationalError:
                    traceback.print_exc()
                    print_msg("Couldn't store graph cache", True)
                return graph
            return load
        return dec
    return cache


class TreeElement(object):
    """ Base class for Intermediate Tree """

    def __init__(self, level=-1, use_id=True):
        self.duration = 0
        self.count = 1
        self.repr = ""
        self.level = level
        self.use_id = use_id
        self.trial_id = 0

    def mean(self):
        """ Mean duration of Tree Node """
        return self.duration / self.count

    def visit(self, visitor):
        """ Visitor pattern """
        name = "visit_{}".format(self.__class__.__name__.lower())
        return getattr(visitor, name)(self)

    def calculate_repr(self):
        """ Calculate Representation of Node """
        pass

    def mix(self, other):
        """ Combine Nodes """
        pass

    def __hash__(self):
        if self.use_id:
            return id(self)
        return hash(self.__repr__())

    def __repr__(self):
        return self.repr

    def to_dict(self, nid):
        """ Convert element to dict """
        return {}


class Single(TreeElement):
    """ Single Node
        Represent an activation or a group of merged activations """
    # pylint: disable=R0902
    # pylint: disable=C0103
    def __init__(self, activation):
        super(Single, self).__init__(level=0, use_id=True)
        self.activation = activation
        self.activations = {activation}
        self.parent = activation['caller_id']
        self.id = activation['id']
        self.line = activation['line']
        self.name = activation['name']
        self.trial_id = activation['trial_id']
        self.repr = "S({0}-{1})".format(self.line, self.name)
        self.finished = bool(activation['finish'])

    @property
    def count(self):
        """ Count activations """
        return sum(1 for a in self.activations)

    @count.setter
    def count(self, value):
        """ Ignore set count """
        pass

    @property
    def duration(self):
        """ Calculate total duration """
        return sum(calculate_duration(a) for a in self.activations
                   if a['finish'] and a['start'])

    @duration.setter
    def duration(self, value):
        """ Ignore set duration """
        pass

    def mix(self, other):
        """ Combine activations """
        self.finished &= other.finished
        self.count += other.count
        self.duration += other.duration
        self.activations = self.activations.union(other.activations)

    def __eq__(self, other):
        return all(fn(self) == fn(other) for fn in
                   [type, lambda x: x.line, lambda x: x.name])

    def name_id(self):
        """ Return name that identifies node """
        return "{0} {1}".format(self.line, self.name)

    def visit(self, visitor):
        return visitor.visit_single(self)

    def to_dict(self, nid):
        return {
            'index': nid,
            'caller_id': self.parent,
            'name': self.name,
            'node': {
                'trial_id': self.trial_id,
                'line': self.line,
                'count': self.count,
                'duration': self.duration,
                'level': self.level,
                'finished': self.finished,
                'info': Info(self),
            },
        }

    def __hash__(self):
        return super(Single, self).__hash__()


class Mixed(TreeElement):
    """ Single Node
        Represent complex Mixed Node """
    # pylint: disable=R0902
    # pylint: disable=C0103
    def __init__(self, activation):
        super(Mixed, self).__init__(level=0, use_id=True)
        self.duration = activation.duration
        self.elements = [activation]
        self.parent = activation.parent
        self.id = activation.id
        self.repr = activation.repr
        self.finished = activation.finished

    @property
    def count(self):
        """ Count activations """
        return sum(e.count for e in self.elements)

    @count.setter
    def count(self, value):
        """ Ignore set count """
        pass

    @property
    def duration(self):
        """ Calculate total duration """
        return sum(e.duration for e in self.elements)

    @duration.setter
    def duration(self, value):
        """ Ignore set duration """
        pass

    @property
    def first(self):
        """ Get first node """
        return next(iter(self.elements))


    def add_element(self, element):
        """ Add node """
        self.finished &= element.finished
        self.elements.append(element)

    def mix(self, other):
        """ Add nodes """
        for element in other.elements:
            self.finished &= element.finished

        self.elements += other.elements
        self.mix_results()

    def mix_results(self):
        """ Combine results """
        it = iter(self.elements)
        initial = next(it)
        for element in it:
            initial.mix(element)

    def __hash__(self):
        return super(Mixed, self).__hash__()


class Group(TreeElement):
    """ Group Node
        Represent a group of called nodes """
    # pylint: disable=R0902
    def __init__(self):
        super(Group, self).__init__(level=0, use_id=True)
        self.nodes = OrderedDict()
        self.edges = OrderedDict()
        self.duration = 0
        self.parent = None
        self.count = 1
        self.repr = ""
        self.finished = True
        self.next_element = None
        self.last = None

    def initialize(self, previous, nnext):
        """ Initialize the group node by two nodes """
        self.nodes[nnext] = Mixed(nnext)
        self.duration = nnext.duration
        self.next_element = nnext
        self.last = nnext
        self.add_subelement(previous)
        self.parent = nnext.parent
        self.finished = previous.finished and nnext.finished
        return self

    def add_subelement(self, previous):
        """ Add new node in sequence to group """
        self.finished &= previous.finished
        nnext, self.next_element = self.next_element, previous
        if not previous in self.edges:
            self.edges[previous] = OrderedCounter()
        if not previous in self.nodes:
            self.nodes[previous] = Mixed(previous)
        else:
            self.nodes[previous].add_element(previous)
        self.edges[previous][nnext] += 1

    def calculate_repr(self):
        result = [
            "[{0}-{1}->{2}]".format(previous, count, nnext)
            for previous, edges in items(self.edges)
            for nnext, count in items(edges)
        ]

        self.repr = "G({0})".format(', '.join(result))

    def __eq__(self, other):
        return all(fn(self) == fn(other) for fn in
                   [type, lambda x: x.edges])

    def mix(self, other):
        for node, value in items(self.nodes):
            value.mix(other.nodes[node])

    def __hash__(self):
        return super(Group, self).__hash__()


class Call(TreeElement):
    """ Call Node
        Represent a call from a Node to a Group """
    # pylint: disable=R0902
    # pylint: disable=C0103
    def __init__(self, caller, called):
        super(Call, self).__init__(level=0, use_id=True)
        self.caller = caller
        self.called = called
        self.called.calculate_repr()
        self.parent = caller.parent
        self.count = 1
        self.id = self.caller.id
        self.duration = self.caller.duration
        self.repr = 'C({0}, {1})'.format(self.caller, self.called)
        self.finished = self.caller.finished

    def __eq__(self, other):
        return all(fn(self) == fn(other) for fn in
                   [type, lambda x: x.caller, lambda x: x.called])

    def mix(self, other):
        self.caller.mix(other.caller)
        self.called.mix(other.called)

    def __hash__(self):
        return super(Call, self).__hash__()


def duration_text(duration, count):
    """ Return duration text """
    return "Total duration: {} microseconds for {} activations".format(
        duration, count)


def mean_text(mean):
    """ Return mean text """
    return "Mean: {} microseconds per activation".format(mean)


def activation_text(activation):
    """ Return single activation text """
    values = activation.arguments
    extra = "(still running)"
    if activation['finish']:
        extra = "to {finish} ({d} microseconds)".format(
            d=calculate_duration(activation), **activation)
    result = [
        "",
        "Activation #{id} from {start} {extra}".format(
            extra=extra, **activation),
    ]
    if values:
        result.append("Arguments: {}".format(
            ", ".join("{}={}".format(value["name"], value["value"])
                      for value in values)))
    if activation['return']:
        result.append("Returned {}".format(activation['return']))
    return result


class Info(object):
    """ Info class
        Represent the information of a node """

    def __init__(self, single):
        self.title = (
            "Trial {trial}<br>"
            "Function <b>{name}</b> called at line {line}").format(
                trial=single.trial_id, name=single.name, line=single.line)
        self.activations = set()
        self.duration = ""
        self.mean = ""
        self.extract_activations(single)
        self.activation_list = None

    def update_by_node(self, node):
        """ Update information by node """
        self.duration = duration_text(node['duration'], node['count'])
        self.mean = mean_text(node['mean'])
        self.activation_list = sorted(self.activations, key=lambda a: a[0])

    def add_activation(self, activation):
        """ Add activation information """
        self.activations.add(
            (datetime.strptime(activation['start'], FORMAT), activation))

    def extract_activations(self, single):
        """ Add activations information """
        for activation in single.activations:
            self.add_activation(activation)

    def __repr__(self):
        result = [self.title, self.duration, self.mean]
        for activation in self.activation_list:
            result += activation_text(activation[1])

        return '<br/>'.join(result)




# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .provider import Provider, row_to_dict
from .database import DatabaseProvider
from .storage import StorageProvider
from .trial import TrialProvider
from .run import RunProvider
from .restore import RestoreProvider


class Persistence(RestoreProvider, DatabaseProvider, TrialProvider,
				  RunProvider, StorageProvider):
	pass

persistence = Persistence()


def get_serialize(arg):
    # ToDo: use arg to select serialize
    #from .serializers import jsonpickle_serializer, jsonpickle_content
    #from .serializers import SimpleSerializer
    # return SimpleSerializer().serialize
    # return jsonpickle_serializer
    # return jsonpickle_content
    return repr


__all__ = [
    b'persistence',
    b'row_to_dict',
    b'get_serializer',
]



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from collections import defaultdict
from ..persistence import row_to_dict, persistence
from ..cross_version import lmap
from ..utils import hashabledict

class FunctionDef(hashabledict):

    @property
    def objects(self):
        if not hasattr(self, '_objects'):
            self._types = defaultdict(list)
            self._objects = lmap(row_to_dict, persistence.load(
                'object', function_def_id=self['id']))

            for obj in self._objects:
                self._types[obj['type']].append(obj)
        return self._objects

    @property
    def types(self):
        self.objects
        return self._types

    @property
    def globals(self):
        return self.types['GLOBAL']

    @property
    def arguments(self):
        return self.types['ARGUMENT']

    @property
    def function_calls(self):
        return self.types['FUNCTION_CALL']



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import sys
from datetime import datetime
from collections import defaultdict

from .data_objects import ObjectStore, FileAccess
from ..persistence import persistence
from ..cross_version import builtins
from ..prov_definition.definition import Definition


class ExecutionProvider(object):

    def __init__(self, metascript):
        # Indicates when activations should be collected
        #   (only after the first call to the script)
        self.enabled = False
        # User files
        self.script = metascript.path
        self.paths = metascript.paths
        # Which function types ('main', 'package' or 'all')
        #   should be considered for the threshold
        self.context = metascript.context
        # How deep we want to go when capturing function activations?
        self.depth_threshold = metascript.depth
        # How deep we want to go beyond our context
        self.non_user_depth_threshold = metascript.non_user_depth
       
        # Object serializer function
        self.serialize = metascript.serialize

        self.metascript = metascript
        self.trial_id = metascript.trial_id
        self.event_map = defaultdict(lambda: self.trace_empty, {})
        self.default_profile = sys.getprofile()
        self.default_trace = sys.gettrace()

    def trace_empty(self, frame, event, arg):
        pass

    def tracer(self, frame, event, arg):
        self.event_map[event](frame, event, arg)

    def store(self, partial=False):
        pass

    def teardown(self):
        self.enabled = False

    def tearup(self):
        pass


class StoreOpenMixin(ExecutionProvider):

    def __init__(self, *args):
        super(StoreOpenMixin, self).__init__(*args)
        persistence.std_open = open
        builtins.open = self.new_open(open)

        # Store provenance
        self.file_accesses = ObjectStore(FileAccess)

    def add_file_access(self, file_access):
        'The class that uses this mixin must override this method'
        pass

    def new_open(self, old_open):
        'Wraps the open buildin function to register file access'
        def open(name, *args, **kwargs):  # @ReservedAssignment
            if self.enabled:
                # Create a file access object with default values
                fid = self.file_accesses.add(name)
                file_access = self.file_accesses[fid]

                if os.path.exists(name):
                    # Read previous content if file exists
                    with old_open(name, 'rb') as f:
                        file_access.content_hash_before = persistence.put(
                            f.read())

                # Update with the informed keyword arguments (mode / buffering)
                file_access.update(kwargs)
                # Update with the informed positional arguments
                if len(args) > 0:
                    file_access.mode = args[0]
                elif len(args) > 1:
                    file_access.buffering = args[1]

                self.add_file_access(file_access)
            return old_open(name, *args, **kwargs)

        return open

    def teardown(self):
        'Restores default open'
        builtins.open = persistence.std_open
        super(StoreOpenMixin, self).teardown()



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import weakref
from collections import defaultdict
from ..cross_version import items


class Model(object):

    __refs__ = defaultdict(list)
    def __init__(self, *args, **kwargs):
        self.__refs__[self.__class__].append(weakref.ref(self))

    def initialize_default(self, kwargs):
        for key, value in items(self.REPLACE):
            if key in kwargs:
                kwargs[value] = kwargs[key]
                del kwargs[key]
        for key, value in items(self.DEFAULT):
            obj = self
            if '.' in key:
                key0, key = key.split('.')
                obj = getattr(self, key0)
            setattr(obj, key, value)
        for key, value in items(kwargs):
            obj = self
            if '.' in key:
                key0, key = key.split('.')
                obj = getattr(self, key0)
            if not hasattr(obj, key):
                setattr(obj, key, value)

    @classmethod
    def get_instances(cls):
        for inst_ref in cls.__refs__[cls]:
            inst = inst_ref()
            if inst is not None:
                yield inst

    @classmethod
    def all_models(cls):
        for c in cls.__refs__:
            for instance in c.get_instances():
                yield instance



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
import pyposast
from datetime import datetime

from ..utils import print_msg, meta_profiler
from ..persistence import persistence
from .function_visitor import FunctionVisitor
from .slicing_visitor import SlicingVisitor
from .definition import Definition
from .utils import FunctionCall, ClassDef, Decorator, Generator, Assert, With
from ..cross_version import cvmap



def visit_ast(metascript, path):
    '''returns a visitor that visited the tree and filled the attributes:
        functions: map of function in the form:
            name -> (arguments, global_vars, calls, code_hash)
        name_refs[path]: map of identifiers in categories Load, Store
        dependencies[path]: map of dependencies
    '''
    with open(path, 'rb') as f:
        code = f.read()

    try:
        tree = pyposast.parse(code, path)
    except SyntaxError:
        print_msg('Syntax error on file {}. Skipping file.'.format(path))
        return None

    visitor = SlicingVisitor(metascript, code, path)
    visitor.result = visitor.visit(tree)
    visitor.extract_disasm()
    visitor.teardown()
    return visitor

@meta_profiler("definition")
def collect_provenance(metascript):
    print_msg('  registering user-defined functions')
    for path in metascript.paths:
        visitor = visit_ast(metascript, path)
        if visitor:
            persistence.store_function_defs(metascript.trial_id,
                                            visitor.functions)
            if metascript.disasm:
                print('------------------------------------------------------')
                print(path)
                print('------------------------------------------------------')
                print('\n'.join(cvmap(repr, visitor.disasm)))
                print('------------------------------------------------------')
            metascript.definition.add_visitor(visitor)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" 'now show' command """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from .command import Command
from ..cross_version import items, values, cvmap
from ..models.trial import Trial
from ..persistence import persistence
from ..utils.functions import wrap
from ..utils.io import print_msg


def print_trial(trial):
    """ Print trial information """
    print(wrap("""\
        Id: {id}
        Inherited Id: {inherited_id}
        Script: {script}
        Code hash: {code_hash}
        Start: {start}
        Finish: {finish}\
        """.format(**trial.info())))


def print_modules(modules):
    """ Print modules """
    output = []
    for module in modules:
        output.append(wrap("""\
            Name: {name}
            Version: {version}
            Path: {path}
            Code hash: {code_hash}\
            """.format(**module), other="\n    "))
    print('\n\n'.join(output))


def print_function_defs(function_defs):
    """ Print function definitions """
    names = lambda lis: cvmap(lambda var: var['name'], lis)
    output = []
    for function_def in function_defs:
        output.append(wrap("""\
            Name: {name}
            Arguments: {arguments}
            Globals: {globals}
            Function calls: {calls}
            Code hash: {code_hash}\
            """.format(arguments=', '.join(names(function_def.arguments)),
                       globals=', '.join(names(function_def.globals)),
                       calls=', '.join(names(function_def.function_calls)),
                       **function_def), other="\n    "))
    print('\n\n'.join(output))


def print_environment(environments):
    """ Print environment variables """
    output = []
    for key, value in items(environments):
        output.append('  {}: {}'.format(key, value))
    print('\n'.join(sorted(output)))


def print_function_activation(trial, activation, level=1):
    """ Print function activation recursively """
    object_values = {'GLOBAL':[], 'ARGUMENT':[]}
    name = {'GLOBAL':'Globals', 'ARGUMENT':'Arguments'}
    for obj in activation.objects:
        object_values[obj['type']].append(
            '{} = {}'.format(obj['name'], obj['value']))
    text = wrap(
        '{line}: {name} ({start} - {finish})'.format(**activation),
        initial='  ' * level)
    indent = text.index(': ') + 2
    print(text)
    for typ, vals in items(object_values):
        print(wrap(
            '{type}: {values}'.format(type=name[typ],
                                      values=', '.join(vals)),
            initial=' ' * indent))
    if activation['return']:
        print(wrap(
            'Return value: {ret}'.format(ret=activation['return']),
            initial=' ' * indent))

    for inner_activation in trial.activations(caller_id=activation['id']):
        print_function_activation(trial, inner_activation, level + 1)


def print_function_activations(trial):
    """ Print function activations """
    for inner_function_activation in trial.activations(caller_id=None):
        print_function_activation(trial, inner_function_activation)


def print_file_accesses(file_accesses):
    """ Print file accesses """
    output = []
    for file_access in file_accesses:
        output.append(wrap("""\
            Name: {name}
            Mode: {mode}
            Buffering: {buffering}
            Content hash before: {content_hash_before}
            Content hash after: {content_hash_after}
            Timestamp: {timestamp}
            Function: {stack}\
            """.format(**file_access)))
    print('\n\n'.join(output))


class Show(Command):
    """ Show the collected provenance of a trial """

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('trial', type=int, nargs='?',
                help='trial id or none for last trial')
        add_arg('-m', '--modules', action='store_true',
                help='shows module dependencies')
        add_arg('-d', '--function-defs', action='store_true',
                help='shows the user-defined functions')
        add_arg('-e', '--environment', action='store_true',
                help='shows the environment conditions')
        add_arg('-a', '--function-activations', action='store_true',
                help='shows function activations')
        add_arg('-f', '--file-accesses', action='store_true',
                help='shows read/write access to files')
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def execute(self, args):
        persistence.connect_existing(args.dir or os.getcwd())
        trial = Trial(trial_id=args.trial, exit=True)
        print_msg('trial information:', True)
        print_trial(trial)

        if args.modules:
            print_msg('this trial depends on the following modules:', True)
            print_modules(trial.modules()[1])

        if args.function_defs:
            print_msg('this trial has the following functions:', True)
            print_function_defs(values(trial.function_defs()))

        if args.environment:
            print_msg('this trial has been executed under the following'
                      ' environment conditions', True)
            print_environment(trial.environment())

        if args.function_activations:
            print_msg('this trial has the following function activation '
                      'graph:', True)
            print_function_activations(trial)

        if args.file_accesses:
            print_msg('this trial accessed the following files:', True)
            print_file_accesses(trial.file_accesses())



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define io utility functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys

from ..cross_version import StringIO, items

LABEL = '[now] '
verbose = False

STDIN = sys.stdin
STDOUT = sys.stdout
STDERR = sys.stderr


class redirect_output(object):

    def __init__(self, outputs=None):
        if outputs is None:
            outputs = ['stdout', 'stderr']
        self.outputs = outputs
        self.old = {}

    def __enter__(self):
        result = []
        for out in self.outputs:
            self.old[out] = getattr(sys, out)
            setattr(sys, out, StringIO())
            result.append(self.old[out])
        return result

    def __exit__(self, exc_type, value, traceback):
        for out, old in items(self.old):
            setattr(sys, out, old)


def print_msg(message, force=False, file=STDOUT):
    """Print message with [now] prefix when in verbose mode"""
    if verbose or force:
        print('{}{}'.format(LABEL, message), file=file)


def print_fn_msg(message, force=False, file=STDOUT):
    """Print lazy message with [now] prefix"""
    if verbose or force:
        print('{}{}'.format(LABEL, message()), file=file)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import textwrap

from datetime import datetime
from ..utils import resource
from .model import Model


RULES = '../resources/rules.pl'


def timestamp(string):
    if not string:
        return -1
    epoch = datetime(1970,1,1)
    time = datetime.strptime(string, '%Y-%m-%d %H:%M:%S.%f')
    return (time - epoch).total_seconds()


class TrialProlog(Model):

    DEFAULT = {}
    REPLACE = {}
    prolog_cli = None

    def __init__(self, trial):
        # TODO: check memory leak?
        self.trial = trial
        self.id = trial.id

    def retract(self):
        list(cls.prolog_cli.query(
            'retract(activation({}, _, _, _, _, _))'.format(self.id)))
        list(cls.prolog_cli.query(
            'retract(access({}, _, _, _, _, _, _, _))'.format(self.id)))
        list(cls.prolog_cli.query(
            'retract(variable({}, _, _, _, _, _))'.format(self.id)))
        list(cls.prolog_cli.query(
            'retract(usage({}, _, _, _, _))'.format(self.id)))
        list(cls.prolog_cli.query(
            'retract(dependency({}, _, _, _))'.format(self.id)))
        list(cls.prolog_cli.query(
            'retract(load_trial({})'.format(self.id)))

    def _trial_activations_fact(self, result):
        result.append(textwrap.dedent("""\
            %
            % FACT: activation(trial_id, id, name, start, finish, caller_activation_id).
            %
            """))

    def export_trial_activations(self, result, with_doc=True):
        if with_doc:
            self._trial_activations_fact(result)
        result.append(":- dynamic(activation/6).")
        for activation in self.trial.activations():
            activation = dict(activation)
            activation['name'] = str(activation['name'])
            activation['start'] = timestamp(activation['start'])
            activation['finish'] = timestamp(activation['finish'])
            if not activation['caller_id']:
                activation['caller_id'] = 'nil'
            result.append(
                'activation('
                    '{trial_id}, {id}, {name!r}, {start:-f}, {finish:-f}, '
                    '{caller_id}).'
                ''.format(**activation))

    def _trial_file_accesses_fact(self, result):
        result.append(textwrap.dedent("""
            %
            % FACT: access(trial_id, id, name, mode, content_hash_before, content_hash_after, timestamp, activation_id).
            %
            """))

    def export_trial_file_accesses(self, result, with_doc=True):
        if with_doc:
            self._trial_file_accesses_fact(result)
        result.append(":- dynamic(access/8).")
        for access in self.trial.file_accesses():
            access = dict(access)
            access['trial_id'] = self.trial.id
            access['name'] = str(access['name'])
            access['mode'] = str(access['mode'])
            access['buffering'] = str(access['buffering'])
            access['content_hash_before'] = str(access['content_hash_before'])
            access['content_hash_after'] = str(access['content_hash_after'])
            access['timestamp'] = timestamp(access['timestamp'])
            result.append(
                'access('
                    '{trial_id}, f{id}, {name!r}, {mode!r}, '
                    '{content_hash_before!r}, {content_hash_after!r}, '
                    '{timestamp:-f}, {function_activation_id}).'
                ''.format(**access))

    def _trial_slicing_variables_fact(self, result):
        result.append(textwrap.dedent("""
            %
            % FACT: variable(trial_id, vid, name, line, value, timestamp).
            %
            """))

    def export_trial_slicing_variables(self, result, with_doc=True):
        if with_doc:
            self._trial_slicing_variables_fact(result)
        result.append(":- dynamic(variable/6).")
        for var in self.trial.slicing_variables():
            var = dict(var)
            var['vid'] = str(var['vid'])
            var['name'] = str(var['name'])
            var['line'] = str(var['line'])
            var['value'] = str(var['value'])
            var['time'] = timestamp(var['time'])
            result.append(
                'variable('
                    '{trial_id}, {vid}, {name!r}, {line}, {value!r}, '
                    '{time:-f}).'
                ''.format(**var))

    def _trial_slicing_usages_fact(self, result):
        result.append(textwrap.dedent("""
            %
            % FACT: usage(trial_id, id, vid, name, line).
            %
            """))

    def export_trial_slicing_usages(self, result, with_doc=True):
        if with_doc:
            self._trial_slicing_usages_fact(result)
        result.append(":- dynamic(usage/5).")
        for usage in self.trial.slicing_usages():
            usage = dict(usage)
            usage['id'] = str(usage['id'])
            usage['vid'] = str(usage['vid'])
            usage['name'] = str(usage['name'])
            usage['line'] = str(usage['line'])
            result.append(
                'usage('
                    '{trial_id}, {id}, {vid}, {name!r}, {line}).'
                ''.format(**usage))

    def _trial_slicing_dependencies_fact(self, result):
        result.append(textwrap.dedent("""
                %
                % FACT: dependency(trial_id, id, dependent, supplier).
                %
            """))

    def export_trial_slicing_dependencies(self, result, with_doc=True):
        if with_doc:
            self._trial_slicing_dependencies_fact(result)
        result.append(":- dynamic(dependency/4).")
        for dep in self.trial.slicing_dependencies():
            dep = dict(dep)
            dep['id'] = str(dep['id'])
            dep['dependent'] = str(dep['dependent'])
            dep['supplier'] = str(dep['supplier'])
            result.append(
                'dependency('
                    '{trial_id}, {id}, {dependent}, {supplier}).'
                ''.format(**dep))

    def export_facts(self, with_doc=True):
        # TODO: export remaining data
        # (now focusing only on activation, file access and slices)
        result = []
        self.export_trial_activations(result, with_doc)
        self.export_trial_file_accesses(result, with_doc)
        self.export_trial_slicing_variables(result, with_doc)
        self.export_trial_slicing_usages(result, with_doc)
        self.export_trial_slicing_dependencies(result, with_doc)
        return result

    def export_text_facts(self):
        return '\n'.join(self.export_facts())

    def export_rules(self, with_facts=False):
        result = []
        if with_facts:
            self._trial_activations_fact(result)
            self._trial_file_accesses_fact(result)
            self._trial_slicing_variables_fact(result)
            self._trial_slicing_usages_fact(result)
            self._trial_slicing_dependencies_fact(result)
        result += resource(RULES, 'UTF-8').split('\n')
        return result

    def load_cli_facts(self):
        self.init_cli()
        load_trial = 'load_trial({})'.format(self.id)
        if not list(self.prolog_cli.query(load_trial)):
            for fact in self.export_facts(with_doc=False):
                self.prolog_cli.assertz(fact[:-1])
            self.prolog_cli.assertz(load_trial)
        load_rules = 'load_rules(1)'
        if not list(self.prolog_cli.query(load_rules)):
            for rule in self.export_rules():
                rule = rule.strip()
                if not rule or rule[0] == '%':
                    continue
                self.prolog_cli.assertz(rule[:-1])
            self.prolog_cli.assertz(load_rules)

    def query(self, query):
        self.load_cli_facts()
        return self.prolog_query(query)

    def retract(self):
        self.retract_trial(self.id)

    def __hash__(self):
        return self.id

    @classmethod
    def init_cli(cls):
        if not cls.prolog_cli:
            from pyswip import Prolog
            cls.prolog_cli = Prolog()
            cls.prolog_cli.assertz('load_trial(0)')
            cls.prolog_cli.assertz('load_rules(0)')

    @classmethod
    def prolog_query(cls, query):
        cls.init_cli()

        cache = set()
        no_cache = set()
        for inst in cls.get_instances():
            (cache if inst.trial.use_cache else no_cache).add(inst)
        for inst in (no_cache - cache):
            inst.retract()


        return cls.prolog_cli.query(query)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define data structures and data access functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from collections import OrderedDict, Counter


def concat_iter(*iters):
    """Concatenate iterators"""
    for iterator in iters:
        for value in iterator:
            yield value


class OrderedCounter(OrderedDict, Counter):
    """OrderedDict with default value 0"""
    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__,
                           OrderedDict(self))
    def __reduce__(self):
        return self.__class__, (OrderedDict(self),)


class HashableDict(dict):
    """Hashable Dict that can be used on sets"""

    def create(self, element):
        """Create hashable element recursively"""
        if isinstance(element, dict):
            return self.__class__(element)
        else:
            return element

    def key(self):
        """Create tuple with elements"""
        return tuple((k, self.create(self[k])) for k in sorted(self))

    def __hash__(self):
        return hash(self.key())

    def __eq__(self, other):
        return self.key() == other.key()


hashabledict = HashableDict



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" 'now list' command """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from .command import Command
from ..persistence import persistence
from ..utils.io import print_msg


class List(Command):
    """ List all trials registered in the current directory """

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def execute(self, args):
        persistence.connect_existing(args.dir or os.getcwd())
        print_msg('trials available in the provenance store:', True)
        for trial in persistence.load('trial'):
            text = '  Trial {id}: {script} {arguments}'.format(**trial)
            indent = text.index(': ') + 2
            print(text)
            print('{indent}with code hash {code_hash}'.format(
                indent=' ' * indent, **trial))
            print('{indent}ran from {start} to {finish}'.format(
                indent=' ' * indent, **trial))
